@@ -21,6 +21,8 @@
 #include <config.h>
 #include <string.h>
 
+#include <glib.h>
+
 #include "lm-debug.h"
 #include "lm-internals.h"
 #include "lm-message-node.h"
@@ -252,7 +254,7 @@ lm_parser_parse (LmParser *parser, const gchar *string)
         }
         
         if (g_markup_parse_context_parse (parser->context, string, 
-                                          strlen (string), NULL)) {
+                                          (gssize)strlen (string), NULL)) {
         } else {
                 g_warning ("Parsing failed\n");
         }
